use crate::cond;

use super::{
    constants::{fix_disliked, Variant},
    contrast_curve::ContrastCurve,
    dynamic_color::{foreground_tone, DynamicColor, ToneDeltaPair, TonePolarity},
    dynamic_scheme::DynamicScheme,
};

macro_rules! dynamic_gen {
    ($name: ident, $palette: expr, $tone: expr) => {
        pub fn $name() -> DynamicColor {
            DynamicColor::new(
                Some(stringify!($name)),
                Box::new($palette),
                Box::new($tone),
                None,
                None,
                None,
                None,
                None,
            )
        }
    };
    ($name: ident, $palette: expr, $tone: expr, $is_background: expr) => {
        pub fn $name() -> DynamicColor {
            DynamicColor::new(
                Some(stringify!($name)),
                Box::new($palette),
                Box::new($tone),
                Some($is_background),
                None,
                None,
                None,
                None,
            )
        }
    };
    ($name: ident, $palette: expr, $tone: expr, $background: expr, $contrast_curve: expr) => {
        pub fn $name() -> DynamicColor {
            DynamicColor::new(
                Some(stringify!($name)),
                Box::new($palette),
                Box::new($tone),
                None,
                Some(Box::new($background)),
                None,
                Some($contrast_curve),
                None,
            )
        }
    };
    ($name: ident, $palette: expr, $tone: expr, $is_background: expr, $background: expr, $secondary_background: expr, $contrast_curve: expr, $tone_delta_pairs: expr) => {
        pub fn $name() -> DynamicColor {
            DynamicColor::new(
                Some(stringify!($name)),
                Box::new($palette),
                Box::new($tone),
                $is_background,
                $background,
                $secondary_background,
                $contrast_curve,
                $tone_delta_pairs,
            )
        }
    };
}

#[inline]
fn is_fidelity(scheme: &DynamicScheme) -> bool {
    scheme.variant == Variant::Fidelity || scheme.variant == Variant::Content
}

#[inline]
fn is_monochrome(scheme: &DynamicScheme) -> bool {
    scheme.variant == Variant::Monochrome
}

fn highest_surface(s: &DynamicScheme) -> DynamicColor {
    cond!(s.is_dark, surface_bright(), surface())
}

dynamic_gen!(
    surface,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 6.0, 98.0),
    true
);
dynamic_gen!(
    surface_dim,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        6.0,
        ContrastCurve::new(87.0, 87.0, 80.0, 75.0).get(s.contrast_level)
    ),
    true
);
dynamic_gen!(
    surface_bright,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(24.0, 24.0, 29.0, 34.0).get(s.contrast_level),
        98.0
    ),
    true
);
dynamic_gen!(
    surface_container_lowest,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(4.0, 4.0, 2.0, 0.0).get(s.contrast_level),
        100.0
    ),
    true
);
dynamic_gen!(
    surface_container_low,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(10.0, 10.0, 11.0, 12.0).get(s.contrast_level),
        ContrastCurve::new(96.0, 96.0, 96.0, 95.0).get(s.contrast_level)
    ),
    true
);
dynamic_gen!(
    surface_container,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(12.0, 12.0, 16.0, 20.0).get(s.contrast_level),
        ContrastCurve::new(94.0, 94.0, 92.0, 90.0).get(s.contrast_level)
    ),
    true
);
dynamic_gen!(
    surface_container_high,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(17.0, 17.0, 21.0, 25.0).get(s.contrast_level),
        ContrastCurve::new(92.0, 92.0, 88.0, 85.0).get(s.contrast_level)
    ),
    true
);
dynamic_gen!(
    surface_container_highest,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(
        s.is_dark,
        ContrastCurve::new(22.0, 22.0, 26.0, 30.0).get(s.contrast_level),
        ContrastCurve::new(90.0, 90.0, 84.0, 80.0).get(s.contrast_level)
    ),
    true
);
dynamic_gen!(
    on_surface,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 90.0, 10.0),
    highest_surface,
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    surface_variant,
    |s: &DynamicScheme| s.neutral_variant_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 30.0, 90.0),
    true
);
dynamic_gen!(
    on_surface_variant,
    |s: &DynamicScheme| s.neutral_variant_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 80.0, 30.0),
    highest_surface,
    ContrastCurve::new(3.0, 4.5, 7.0, 11.0)
);
dynamic_gen!(
    inverse_surface,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 20.0, 95.0)
);
dynamic_gen!(
    inverse_on_surface,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 20.0, 95.0),
    |_| inverse_surface(),
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    outline,
    |s: &DynamicScheme| s.neutral_variant_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 60.0, 50.0),
    highest_surface,
    ContrastCurve::new(1.5, 3.0, 4.5, 7.0)
);
dynamic_gen!(
    outline_variant,
    |s: &DynamicScheme| s.neutral_variant_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 30.0, 80.0),
    highest_surface,
    ContrastCurve::new(1.0, 1.0, 3.0, 4.5)
);
dynamic_gen!(
    shadow,
    |s: &DynamicScheme| s.neutral_palette.clone(),
    |_| 0.0
);
dynamic_gen!(scrim, |s: &DynamicScheme| s.neutral_palette.clone(), |_| {
    0.0
});
dynamic_gen!(
    primary,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            cond!(s.is_dark, 100.0, 0.0)
        } else {
            cond!(s.is_dark, 80.0, 40.0)
        }
    },
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 7.0)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: primary_container(),
        role_b: primary(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_primary,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            cond!(s.is_dark, 10.0, 90.0)
        } else {
            cond!(s.is_dark, 20.0, 100.0)
        }
    },
    |_| primary(),
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    primary_container,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| {
        if is_fidelity(s) {
            return s.source_color.l;
        }
        if is_monochrome(s) {
            cond!(s.is_dark, 85.0, 25.0)
        } else {
            cond!(s.is_dark, 30.0, 90.0)
        }
    },
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: primary_container(),
        role_b: primary(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_primary_container,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| {
        if is_fidelity(s) {
            return foreground_tone(primary_container().get_tone(s), 4.5);
        }
        if is_monochrome(s) {
            cond!(s.is_dark, 0.0, 100.0)
        } else {
            cond!(s.is_dark, 90.0, 30.0)
        }
    },
    |_| primary_container(),
    ContrastCurve::new(3.0, 4.5, 7.0, 11.0)
);
dynamic_gen!(
    inverse_primary,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 40.0, 80.0),
    |_| inverse_surface(),
    ContrastCurve::new(3.0, 4.5, 7.0, 7.0)
);
dynamic_gen!(
    secondary,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 80.0, 40.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 7.0)),
    None
);
dynamic_gen!(
    on_secondary,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            cond!(s.is_dark, 10.0, 100.0)
        } else {
            cond!(s.is_dark, 20.0, 100.0)
        }
    },
    |_| secondary(),
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    secondary_container,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| {
        let initial_tone = cond!(s.is_dark, 30.0, 90.0);
        if is_monochrome(s) {
            return cond!(s.is_dark, 30.0, 85.0);
        }
        initial_tone
    },
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: secondary_container(),
        role_b: secondary(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_secondary_container,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            return cond!(s.is_dark, 90.0, 10.0);
        }
        if !is_fidelity(s) {
            return cond!(s.is_dark, 90.0, 30.0);
        }
        foreground_tone(secondary_container().get_tone(s), 4.5)
    },
    |_| secondary_container(),
    ContrastCurve::new(3.0, 4.5, 7.0, 11.0)
);
dynamic_gen!(
    tertiary,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            return cond!(s.is_dark, 90.0, 25.0);
        }
        cond!(s.is_dark, 80.0, 40.0)
    },
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 7.0)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: tertiary_container(),
        role_b: tertiary(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_tertiary,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            return cond!(s.is_dark, 10.0, 90.0);
        }
        cond!(s.is_dark, 20.0, 100.0)
    },
    |_| tertiary(),
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    tertiary_container,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            return cond!(s.is_dark, 60.0, 49.0);
        }
        if !is_fidelity(s) {
            return cond!(s.is_dark, 30.0, 90.0);
        }
        let proposed_lch = s.tertiary_palette.tone(s.source_color.l);
        fix_disliked(&proposed_lch).l
    },
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: tertiary_container(),
        role_b: tertiary(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_tertiary_container,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            return cond!(s.is_dark, 0.0, 100.0);
        }
        if !is_fidelity(s) {
            return cond!(s.is_dark, 90.0, 30.0);
        }
        foreground_tone(tertiary_container().get_tone(s), 4.5)
    },
    |_| tertiary_container(),
    ContrastCurve::new(3.0, 4.5, 7.0, 11.0)
);
dynamic_gen!(
    error,
    |s: &DynamicScheme| s.error_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 80.0, 40.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 7.0)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: error_container(),
        role_b: error(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_error,
    |s: &DynamicScheme| s.error_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 20.0, 100.0),
    |_| error(),
    ContrastCurve::new(4.5, 7.0, 11.0, 21.0)
);
dynamic_gen!(
    error_container,
    |s: &DynamicScheme| s.error_palette.clone(),
    |s: &DynamicScheme| cond!(s.is_dark, 30.0, 90.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: error_container(),
        role_b: error(),
        delta: 10.0,
        polarity: TonePolarity::Nearer,
        togather: false,
    }))
);
dynamic_gen!(
    on_error_container,
    |s: &DynamicScheme| s.error_palette.clone(),
    |s: &DynamicScheme| {
        if is_monochrome(s) {
            cond!(s.is_dark, 90.0, 10.0)
        } else {
            cond!(s.is_dark, 90.0, 30.0)
        }
    },
    |_| error_container(),
    ContrastCurve::new(3.0, 4.5, 7.0, 11.0)
);
dynamic_gen!(
    primary_fixed,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 40.0, 90.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: primary_fixed(),
        role_b: primary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    primary_fixed_dim,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 30.0, 80.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: primary_fixed(),
        role_b: primary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    on_primary_fixed,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 100.0, 10.0),
    None,
    Some(Box::new(|_| primary_fixed_dim())),
    Some(Box::new(|_| primary_fixed())),
    Some(ContrastCurve::new(4.5, 7.0, 11.0, 21.0)),
    None
);
dynamic_gen!(
    on_primary_fixed_variant,
    |s: &DynamicScheme| s.primary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 90.0, 30.0),
    None,
    Some(Box::new(|_| primary_fixed_dim())),
    Some(Box::new(|_| primary_fixed())),
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
    None
);
dynamic_gen!(
    secondary_fixed,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 80.0, 90.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: secondary_fixed(),
        role_b: secondary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    secondary_fixed_dim,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 70.0, 80.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: secondary_fixed(),
        role_b: secondary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    on_secondary_fixed,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |_| 10.0,
    None,
    Some(Box::new(|_| secondary_fixed_dim())),
    Some(Box::new(|_| secondary_fixed())),
    Some(ContrastCurve::new(4.5, 7.0, 11.0, 21.0)),
    None
);
dynamic_gen!(
    on_secondary_fixed_variant,
    |s: &DynamicScheme| s.secondary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 25.0, 30.0),
    None,
    Some(Box::new(|_| secondary_fixed_dim())),
    Some(Box::new(|_| secondary_fixed())),
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
    None
);
dynamic_gen!(
    tertiary_fixed,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 40.0, 90.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: tertiary_fixed(),
        role_b: tertiary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    tertiary_fixed_dim,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 30.0, 80.0),
    Some(true),
    Some(Box::new(highest_surface)),
    None,
    Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
    Some(Box::new(|_| ToneDeltaPair {
        role_a: tertiary_fixed(),
        role_b: tertiary_fixed_dim(),
        delta: 10.0,
        polarity: TonePolarity::Lighter,
        togather: true,
    }))
);
dynamic_gen!(
    on_tertiary_fixed,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 100.0, 10.0),
    None,
    Some(Box::new(|_| tertiary_fixed_dim())),
    Some(Box::new(|_| tertiary_fixed())),
    Some(ContrastCurve::new(4.5, 7.0, 11.0, 21.0)),
    None
);
dynamic_gen!(
    on_tertiary_fixed_variant,
    |s: &DynamicScheme| s.tertiary_palette.clone(),
    |s: &DynamicScheme| cond!(is_monochrome(s), 90.0, 30.0),
    None,
    Some(Box::new(|_| tertiary_fixed_dim())),
    Some(Box::new(|_| tertiary_fixed())),
    Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
    None
);

pub fn custom(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("{}", name)),
        {
            let name = name.clone();
            Box::new(move |s: &DynamicScheme| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s: &DynamicScheme| {
            if is_monochrome(s) {
                cond!(s.is_dark, 100.0, 0.0)
            } else {
                cond!(s.is_dark, 80.0, 40.0)
            }
        }),
        Some(true),
        Some(Box::new(highest_surface)),
        None,
        Some(ContrastCurve::new(3.0, 4.5, 7.0, 7.0)),
        {
            let name = name.clone();
            Some(Box::new(move |_| ToneDeltaPair {
                role_a: custom_container(String::from(&name)),
                role_b: custom(String::from(&name)),
                delta: 10.0,
                polarity: TonePolarity::Nearer,
                togather: false,
            }))
        },
    )
}

pub fn on_custom(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("on_{}", name)),
        {
            let name = name.clone();
            Box::new(move |s| {
                let custom_palette = s.custom_palettes.get(&name).unwrap().clone();
                custom_palette
            })
        },
        Box::new(|s: &DynamicScheme| {
            if is_monochrome(s) {
                cond!(s.is_dark, 10.0, 90.0)
            } else {
                cond!(s.is_dark, 20.0, 100.0)
            }
        }),
        None,
        {
            let name = name.clone();
            Some(Box::new(move |_| custom(String::from(&name))))
        },
        None,
        Some(ContrastCurve::new(4.5, 7.0, 11.0, 21.0)),
        None,
    )
}

pub fn custom_container(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("{}_container", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| {
            if is_fidelity(s) {
                return s.source_color.l;
            }
            if is_monochrome(s) {
                cond!(s.is_dark, 85.0, 25.0)
            } else {
                cond!(s.is_dark, 30.0, 90.0)
            }
        }),
        Some(true),
        Some(Box::new(highest_surface)),
        None,
        Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
        {
            let name = name.clone();
            Some(Box::new(move |_| ToneDeltaPair {
                role_a: custom_container(String::from(&name)),
                role_b: custom(String::from(&name)),
                delta: 10.0,
                polarity: TonePolarity::Nearer,
                togather: false,
            }))
        },
    )
}

pub fn on_custom_container(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("on_{}_container", name)),
        {
            let name = name.clone();
            Box::new(move |s| {
                let custom_palette = s.custom_palettes.get(&name).unwrap().clone();
                custom_palette
            })
        },
        {
            let name = name.clone();
            Box::new(move |s| {
                if is_fidelity(s) {
                    return foreground_tone(custom_container(String::from(&name)).get_tone(s), 4.5);
                }
                if is_monochrome(s) {
                    cond!(s.is_dark, 0.0, 100.0)
                } else {
                    cond!(s.is_dark, 90.0, 30.0)
                }
            })
        },
        None,
        {
            let name = name.clone();
            Some(Box::new(move |_| custom_container(String::from(&name))))
        },
        None,
        Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
        None,
    )
}

pub fn inverse_custom(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("inverse_{}", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| cond!(s.is_dark, 20.0, 95.0)),
        None,
        Some(Box::new(|_| inverse_surface())),
        None,
        Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
        None,
    )
}

pub fn custom_fixed(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("{}_fixed", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| cond!(is_monochrome(s), 40.0, 90.0)),
        Some(true),
        Some(Box::new(highest_surface)),
        None,
        Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
        {
            let name = name.clone();
            Some(Box::new(move |_| ToneDeltaPair {
                role_a: custom_fixed(String::from(&name)),
                role_b: custom_fixed_dim(String::from(&name)),
                delta: 10.0,
                polarity: TonePolarity::Lighter,
                togather: true,
            }))
        },
    )
}

pub fn custom_fixed_dim(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("{}_fixed_dim", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| cond!(is_monochrome(s), 30.0, 80.0)),
        Some(true),
        Some(Box::new(highest_surface)),
        None,
        Some(ContrastCurve::new(1.0, 1.0, 3.0, 4.5)),
        {
            let name = name.clone();
            Some(Box::new(move |_| ToneDeltaPair {
                role_a: custom_fixed(String::from(&name)),
                role_b: custom_fixed_dim(String::from(&name)),
                delta: 10.0,
                polarity: TonePolarity::Lighter,
                togather: true,
            }))
        },
    )
}

pub fn on_custom_fixed(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("on_{}_fixed", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| cond!(is_monochrome(s), 100.0, 10.0)),
        None,
        {
            let name = name.clone();
            Some(Box::new(move |_| custom_fixed_dim(String::from(&name))))
        },
        Some(Box::new(move |_| custom_fixed(String::from(&name)))),
        Some(ContrastCurve::new(4.5, 7.0, 11.0, 21.0)),
        None,
    )
}

pub fn on_custom_fixed_variant(name: String) -> DynamicColor {
    DynamicColor::new(
        Some(&format!("on_{}_fixed_variant", name)),
        {
            let name = name.clone();
            Box::new(move |s| s.custom_palettes.get(&name).unwrap().clone())
        },
        Box::new(|s| cond!(is_monochrome(s), 90.0, 30.0)),
        None,
        {
            let name = name.clone();
            Some(Box::new(move |_| custom_fixed_dim(String::from(&name))))
        },
        Some(Box::new(move |_| custom_fixed(String::from(&name)))),
        Some(ContrastCurve::new(3.0, 4.5, 7.0, 11.0)),
        None,
    )
}
