use std::{collections::HashMap, str::FromStr};

use dynamic_scheme::DynamicScheme;
use material_colors::*;
use palette::{IntoColor, Lch, Srgb};
use wasm_bindgen::{prelude::wasm_bindgen, JsValue};

use crate::{cond, convert::map_lch_to_srgb_hex, errors};

use super::material_design_3::{M3BaselineColors, M3ColorSet, M3SurfaceSet};
pub use constants::Variant;

mod constants;
mod contrast;
mod contrast_curve;
mod dynamic_color;
mod dynamic_scheme;
mod material_colors;

#[wasm_bindgen]
pub fn material_design_3_dynamic_variant() -> Result<JsValue, String> {
    let variants = enum_iterator::all::<constants::Variant>()
        .map(|variant| {
            serde_json::json!({
                "label": variant.label(),
                "value": variant as u8,
            })
        })
        .collect::<Vec<_>>();

    Ok(serde_wasm_bindgen::to_value(&variants).map_err(|e| e.to_string())?)
}

fn build_primary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&primary().get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_primary().get_lch(scheme)),
        container: map_lch_to_srgb_hex(&primary_container().get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_primary_container().get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&primary_fixed().get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&on_primary_fixed().get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&on_primary_fixed_variant().get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&primary_fixed_dim().get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&inverse_primary().get_lch(scheme)),
    }
}

fn build_secondary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&secondary().get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_secondary().get_lch(scheme)),
        container: map_lch_to_srgb_hex(&secondary_container().get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_secondary_container().get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&secondary_fixed().get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&on_secondary_fixed().get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&on_secondary_fixed_variant().get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&secondary_fixed_dim().get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.secondary_palette),
            M3ColorSet::new_light_set(&scheme.secondary_palette)
        )
    }
}

fn build_tertiary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&tertiary().get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_tertiary().get_lch(scheme)),
        container: map_lch_to_srgb_hex(&tertiary_container().get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_tertiary_container().get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&tertiary_fixed().get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&on_tertiary_fixed().get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&on_tertiary_fixed_variant().get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&tertiary_fixed_dim().get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.tertiary_palette),
            M3ColorSet::new_light_set(&scheme.tertiary_palette)
        )
    }
}

fn build_surface_color_set(scheme: &DynamicScheme) -> M3SurfaceSet {
    M3SurfaceSet {
        root: map_lch_to_srgb_hex(&surface().get_lch(scheme)),
        dim: map_lch_to_srgb_hex(&surface_dim().get_lch(scheme)),
        bright: map_lch_to_srgb_hex(&surface_bright().get_lch(scheme)),
        variant: map_lch_to_srgb_hex(&surface_variant().get_lch(scheme)),
        container: map_lch_to_srgb_hex(&surface_container().get_lch(scheme)),
        container_lowest: map_lch_to_srgb_hex(&surface_container_lowest().get_lch(scheme)),
        container_low: map_lch_to_srgb_hex(&surface_container_low().get_lch(scheme)),
        container_high: map_lch_to_srgb_hex(&surface_container_high().get_lch(scheme)),
        container_highest: map_lch_to_srgb_hex(&surface_container_highest().get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_surface().get_lch(scheme)),
        on_root_variant: map_lch_to_srgb_hex(&on_surface_variant().get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&inverse_surface().get_lch(scheme)),
        on_inverse: map_lch_to_srgb_hex(&inverse_on_surface().get_lch(scheme)),
    }
}

fn build_error_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&error().get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_error().get_lch(scheme)),
        container: map_lch_to_srgb_hex(&error_container().get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_error_container().get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.error_palette),
            M3ColorSet::new_light_set(&scheme.error_palette)
        )
    }
}

fn build_custom_color_set(scheme: &DynamicScheme, name: String) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&custom(name.clone()).get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_custom(name.clone()).get_lch(scheme)),
        container: map_lch_to_srgb_hex(&custom_container(name.clone()).get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_custom_container(name.clone()).get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&custom_fixed(name.clone()).get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&on_custom_fixed(name.clone()).get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&on_custom_fixed_variant(name.clone()).get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&custom_fixed_dim(name.clone()).get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&inverse_custom(name.clone()).get_lch(scheme)),
    }
}

pub fn build_baseline(scheme: &DynamicScheme) -> M3BaselineColors {
    M3BaselineColors::full_custom(
        build_primary_color_set(scheme),
        build_secondary_color_set(scheme),
        build_tertiary_color_set(scheme),
        build_error_color_set(scheme),
        build_surface_color_set(scheme),
        map_lch_to_srgb_hex(&outline().get_lch(scheme)),
        map_lch_to_srgb_hex(&outline_variant().get_lch(scheme)),
        map_lch_to_srgb_hex(&scrim().get_lch(scheme)),
        map_lch_to_srgb_hex(&shadow().get_lch(scheme)),
        scheme
            .custom_palettes
            .keys()
            .map(|name| (name.clone(), build_custom_color_set(scheme, name.clone())))
            .collect(),
        scheme.is_dark,
    )
}

pub fn build_dynamic_scheme(
    source_color: &str,
    error_color: Option<String>,
    custom_colors: HashMap<String, String>,
    variant: constants::Variant,
    contrast_level: f32,
    is_dark: bool,
    harmonize_customs: bool,
) -> Result<DynamicScheme, errors::ColorError> {
    let source_color = Srgb::from_str(source_color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(source_color.to_string()))?
        .into_format::<f32>()
        .into_color();
    let error_color = error_color
        .map(|color| Srgb::from_str(&color))
        .transpose()
        .map_err(|_| errors::ColorError::UnrecogniazedRGB("error color".to_string()))?
        .map(|color| color.into_format::<f32>().into_color());
    let custom_colors = custom_colors
        .into_iter()
        .map(|(name, color)| {
            let color = Srgb::from_str(&color)
                .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.clone()))?
                .into_format::<f32>()
                .into_color();
            Ok((name, color))
        })
        .collect::<Result<HashMap<String, Lch>, errors::ColorError>>()?;
    Ok(DynamicScheme::new(
        source_color,
        error_color,
        custom_colors,
        variant,
        contrast_level,
        is_dark,
        harmonize_customs,
    ))
}
